local siloClient = require("silo.client")

local ui = {
  keys = {},

  search = {
    value = "",
    focused = true
  },

  withdraw = {
    focused = false,
    finished = false,
    sent = false,
    value = ""
  },

  pickup = {
    requested = false,
    running = false
  },

  scroll = 0,
  selectedItem = nil
}

local SCROLL_MARGIN = 4

local function updateInputMap(event, key)
  if event == "key" then
    ui.keys[key] = true
  end

  if event == "key_up" then
    ui.keys[key] = false
  end
end

local function getItemEntries(itemCache)
  local itemEntries = {}
  for i, item in pairs(itemCache) do
    if item.count ~= 0 then
      if string.match(string.lower(item.meta.displayName),
          string.lower(ui.search.value)) then
        itemEntries[#itemEntries + 1] = item
      end
    end
  end

  table.sort(itemEntries, function(i1, i2) return i1.count > i2.count end)

  return itemEntries
end

local function selectedIndex(itemEntries)
  for i, item in pairs(itemEntries) do
    if item.hash == ui.selectedItem then
      return i
    end
  end

  return 0
end

local function handleKeyDown(itemEntries, key)
  local controlHeld = ui.keys[keys.leftCtrl] or ui.keys[keys.rightCtrl]

  if key == keys.backspace then
    if ui.search.focused then
      ui.search.value = string.sub(ui.search.value, 1, string.len(ui.search.value) - 1)
    end

    if ui.withdraw.focused then
      ui.withdraw.value = string.sub(ui.withdraw.value, 1, string.len(ui.withdraw.value) - 1)
    end
  end

  if key == keys.c and controlHeld then
    if ui.search.focused then
      ui.search.value = ""
    end

    if ui.withdraw.focused then
      ui.withdraw.value = ""
      ui.withdraw.focused = false
      ui.search.focused = true
    end
  end


  if ui.withdraw.focused then
    if key == keys.enter then
      ui.withdraw.finished = true
    end
  elseif ui.search.focused then
    if key == keys.d and controlHeld then
      ui.pickup.requested = true
    end

    if key == keys.up or key == keys.down then
      if key == keys.up then
        local item = itemEntries[selectedIndex(itemEntries) - 1]
        if item ~= nil then
          ui.selectedItem = item.hash
        end
      end

      if key == keys.down then
        local item = itemEntries[selectedIndex(itemEntries) + 1]
        if item ~= nil then
          ui.selectedItem = item.hash
        end
      end
    end

    if key == keys.enter and selectedIndex(itemEntries) ~= 0 then
      ui.search.focused = false
      ui.withdraw = {
        focused = true,
        finished = false,
        sent = false,
        value = ""
      }
    end
    -- TODO: Scrolling
  end
end

local function handleCharacterTyped(char)
  if ui.search.focused then
    ui.search.value = ui.search.value .. char
  end

  if ui.withdraw.focused and not ui.withdraw.finished then
    ui.withdraw.value = ui.withdraw.value .. char
    ui.withdraw.error = nil
  end
end

local function renderItemEntries(itemEntries, width, height)
  if selectedIndex(itemEntries) > (height - (2 + SCROLL_MARGIN)) then
    ui.scroll = (height - (2 + SCROLL_MARGIN)) - selectedIndex(itemEntries)
  else
    ui.scroll = 0
  end

  term.setCursorPos(1, 1)
  term.setTextColour(colours.lightGrey)
  term.write("> ")
  term.setTextColour(colours.white)
  term.write(ui.search.value)

  term.setCursorPos(1, 2)
  term.setTextColour(colours.lightGrey)
  for i = 1, width do
    term.write("=")
  end

  term.setTextColour(colours.white)
  for row = 3, height do
    local i = row - 2 - ui.scroll

    local item = itemEntries[i]
    if item == nil then break end

    term.setCursorPos(1, row)
    term.setTextColour(colours.yellow)

    if item.hash == ui.selectedItem then
      term.write("> ")
    else
      term.write("  ")
    end
    term.setTextColour(colours.white)

    term.write(item.meta.displayName)
    local count = tostring(item.count)
    term.setCursorPos(width - string.len(count), row)
    term.write(count)
  end

  term.setTextColour(colours.white)
  term.setCursorBlink(true)
  term.setCursorPos(3 + string.len(ui.search.value), 1)
end

local function renderWithdrawalScreen(item, width, height)
  term.setCursorPos(1, 1)
  term.setTextColour(colours.lightGrey)
  term.write("Withdraw: ")
  term.setTextColour(colours.white)
  term.write(item.meta.displayName)

  term.setCursorPos(1, 2)
  term.setTextColour(colours.lightGrey)

  local amountPrompt = "Amount (" .. math.min(item.count, item.meta.maxCount) .. "): "
  term.write(amountPrompt)

  if ui.withdraw.error then
    term.setTextColour(colours.red)
  else
    term.setTextColour(colours.white)
  end
  term.write(ui.withdraw.value)

  term.setCursorPos(1, 3)
  term.setTextColour(colours.lightGrey)
  term.write("Available: ")
  term.setTextColour(colours.white)
  term.write(item.count)

  term.setTextColour(colours.lightGrey)
  if item.meta.enchantments then
    for i = 1, #item.meta.enchantments do
      term.setCursorPos(1, 5 + i)
      term.write(item.meta.enchantments[i].fullName)
    end
  end

  term.setCursorPos(string.len(amountPrompt) + 1 + string.len(ui.withdraw.value), 2)
  term.setCursorBlink(not ui.withdraw.finished)
end

siloClient("storage_terminal", function(ctx)
  local config = ctx.config
    :group("storage_terminal", "Storage terminal options")
    :define("pickup_chest", "The container to use this storage terminal with", "minecraft:ironchest_crystal_X")
    :get()

  local protocol = ctx:require("silo.protocol")
  local messaging = ctx:require("silo.messaging")
  local items = ctx:require("silo.items")

  ctx:addTask(function()
    while true do
      local event = table.pack(os.pullEvent())
      local itemEntries = getItemEntries(items.itemCache)

      if selectedIndex(itemEntries) == 0 then
        ui.selectedItem = (itemEntries[1] or {}).hash
      end

      if event[1] == "key" or event[1] == "key_up" then
        updateInputMap(event[1], event[2])

        if event[1] == "key" then
          handleKeyDown(itemEntries, event[2])
        end
      end

      if event[1] == "char" then
        handleCharacterTyped(event[2])
      end

      if ui.pickup.requested and not ui.pickup.running then
        ui.pickup.running = true

        messaging:sendDirect(ctx.server, {
          type = "storage.deposit_items",
          chests = { config["pickup_chest"] }
        }, function(...)
          ui.pickup.requested = false
          ui.pickup.running = false
        end)
      end

      local width, height = term.getSize()
      term.setBackgroundColour(colours.black)
      term.clear()

      if ui.search.focused then
        itemEntries = getItemEntries(items.itemCache)
        renderItemEntries(itemEntries, width, height)
      end

      if ui.withdraw.focused then
        local item = items:getItem(ui.selectedItem)
        if item == nil then
          ui.withdraw.finished = false
          ui.withdraw.focused = false
          ui.search.focused = true
        else
          renderWithdrawalScreen(item, width, height)
        end

        if ui.withdraw.finished then
          local amount

          if ui.withdraw.value == "" then
            amount = math.min(items:getItem(ui.selectedItem).count,
              items:getItem(ui.selectedItem).meta.maxCount)
          else
            amount = tonumber(ui.withdraw.value)
          end

          if not amount then
            ui.withdraw.finished = false
            ui.withdraw.error = true
          elseif not ui.withdraw.sent then
            ui.withdraw.sent = true

            if amount < 0 or amount > items:getItem(ui.selectedItem).count then
              ui.withdraw.error = true
              ui.withdraw.finished = false
            else
              messaging:sendDirect(ctx.server, {
                type = "storage.withdraw_items",
                to = config["pickup_chest"],
                items = {
                  [ui.selectedItem] = amount
                }
              }, function(...)
                ui.withdraw.focused = false
                ui.search.focused = true
              end)
            end
            -- TODO: Ask the server to withdraw the item we want
          end
        end
      end
    end
  end)

  ctx:run()
end)
